import React, { PropTypes } from 'react';
import ReactDOM from 'react-dom';
import PureRenderMixin from 'react-addons-pure-render-mixin';
import onClickOutside from 'react-onclickoutside';
import invariant from 'invariant';
import moment from 'moment';

import * as lang from '../../utils/lang';
import FoundationTether from '../misc/FoundationTether';
import isNullOrEmpty from '../../utils/isNullOrEmpty';
import isHandheld from '../../utils/isHandheld';
import toTimestamp from '../../utils/toTimestamp';

import DatepickerCalendar from './DatepickerCalendar';
import DatepickerControl from './DatepickerControl';
import isVisibleInCurrentMonth from './utils/isVisibleInCurrentMonth';


export default ({
    block,
    displayName,
    uiPropDefaultType = PropTypes.string,
    uiPropDefaultValue = '',
    uiPropGetValue = (ui) => ui,
}) => {

    invariant(block !== undefined, 'Missing purebem block.');
    invariant(!isNullOrEmpty(displayName), 'Missing display name.');

    // These are extracted for better testability because I prefer to select based on
    // display name instead of CSS class.
    const Calendar = DatepickerCalendar({ block, displayName: `${displayName}Calendar` });
    const Control = DatepickerControl({ block, displayName: `${displayName}Control` });

    return onClickOutside(React.createClass({

        displayName,

        mixins: [PureRenderMixin],

        propTypes: {
            currentMonth: PropTypes.number,
            disabled: PropTypes.bool,
            emptyText: PropTypes.string,
            expanded: PropTypes.bool,
            focus: PropTypes.bool,
            hasGrayedOutPast: PropTypes.bool,
            highlightedIndex: PropTypes.number,
            isDisabledDate: PropTypes.func,
            onChange: PropTypes.func,
            onCollapse: PropTypes.func,
            onExpand: PropTypes.func,
            onHighlightedIndexChange: PropTypes.func,
            tetherTo: PropTypes.string,
            ui: uiPropDefaultType,
            value: PropTypes.number,
        },

        getDefaultProps() {
            const timestamp = +new Date();
            return {
                currentMonth: timestamp,
                disabled: false,
                emptyText: lang.pgettext('foundation-datepicker', 'Label--no-date-yet'),
                expanded: false,
                focus: false,
                hasGrayedOutPast: false,
                highlightedIndex: timestamp,
                isDisabledDate: (/* value */) => false,
                onChange: (/* currentMonth, highlightedIndex, value */) => {},
                onCollapse: (/* currentMonth, highlightedIndex */) => {},
                onExpand: () => {},
                onHighlightedIndexChange: (/* currentMonth, highlightedIndex */) => {},
                tetherTo: 'content',
                ui: uiPropDefaultValue,
                value: null,
            };
        },

        componentDidMount() {
            if (this.props.focus) {
                this.focus();
            }
        },

        handleClickOutside() {
            if (this.props.expanded) {
                this.collapse();
            }
        },

        onCollapse() {
            this.collapse();
        },

        onChange(value) {
            this.change(value);
        },

        onKeyDown(evt) {
            let date, newCurrentMonth;
            const currentMonth = moment(this.props.currentMonth);

            switch (evt.which) {
                case 9: // tab
                    if (this.props.expanded) {
                        // don't tab to next form control when expanded.
                        evt.stopPropagation();
                        evt.preventDefault();

                        if (evt.shiftKey) {
                            this.moveToPreviousMonth();
                        } else {
                            this.moveToNextMonth();
                        }
                    }

                    break;
                case 13: // enter
                case 32: // space
                    evt.stopPropagation();
                    evt.preventDefault();

                    if (this.props.expanded) {
                        this.change(moment(this.props.highlightedIndex));
                    } else {
                        this.expand();
                    }

                    break;
                case 27: // escape
                    if (this.props.expanded) {
                        this.collapse();
                    }

                    break;
                case 37: // left
                    evt.preventDefault();

                    if (this.props.expanded) {
                        date = moment(this.props.highlightedIndex).subtract(1, 'days');
                        newCurrentMonth = isVisibleInCurrentMonth(currentMonth, date)
                            ? currentMonth
                            : date.clone().startOf('month');

                        this.changeHighlightedIndex(newCurrentMonth, date);
                    }

                    break;
                case 38: // up
                    evt.preventDefault();

                    if (this.props.expanded) {
                        date = moment(this.props.highlightedIndex).subtract(7, 'days');
                        newCurrentMonth = isVisibleInCurrentMonth(currentMonth, date)
                            ? currentMonth
                            : date.clone().startOf('month');

                        this.changeHighlightedIndex(newCurrentMonth, date);
                    }

                    break;
                case 39: // right
                    evt.preventDefault();

                    if (this.props.expanded) {
                        date = moment(this.props.highlightedIndex).add(1, 'days');
                        newCurrentMonth = isVisibleInCurrentMonth(currentMonth, date)
                            ? currentMonth
                            : date.clone().startOf('month');

                        this.changeHighlightedIndex(newCurrentMonth, date);
                    }

                    break;
                case 40: // down
                    evt.preventDefault();

                    if (this.props.expanded) {
                        date = moment(this.props.highlightedIndex).add(7, 'days');
                        newCurrentMonth = isVisibleInCurrentMonth(currentMonth, date)
                            ? currentMonth
                            : date.clone().startOf('month');

                        this.changeHighlightedIndex(newCurrentMonth, date);
                    } else {
                        this.expand();
                    }

                    break;
            }
        },

        onNextMonthClick() {
            this.moveToNextMonth();
        },

        onExpand() {
            this.expand();
        },

        onPreviousMonthClick() {
            this.moveToPreviousMonth();
        },

        onToggle() {
            const { disabled, expanded } = this.props;
            if (disabled) {
                return;
            }

            expanded ? this.collapse() : this.expand();
        },

        onTouchableChange(evt) {
            if (evt.target.value === '') {
                this.change(null);
            } else {
                const date = moment(evt.target.value, 'YYYY-MM-DD');
                this.change(date);
            }
        },

        change(value) {
            const { disabled, onChange } = this.props;
            if (disabled) {
                return;
            }

            const now = moment();
            const newValue = value === null ? now : value;

            if (this.isDisabledDate(newValue)) {
                return;
            }

            if (value === null) {
                onChange(toTimestamp(now.clone().startOf('month')), toTimestamp(now), toTimestamp(newValue));
            } else if (value.isValid()) {
                onChange(toTimestamp(value.clone().startOf('month')), toTimestamp(value), toTimestamp(newValue));
            }
        },

        changeHighlightedIndex(currentMonth, highlightedIndex, alwaysChangeMonth = false) {
            const { onHighlightedIndexChange } = this.props;

            const disabled = this.isDisabledDate(highlightedIndex);
            if (!alwaysChangeMonth && disabled) {
                return;
            }

            const endOfCurrentMonth = currentMonth.clone().endOf('month');
            const hi = disabled
                ? (!this.isDisabledDate(endOfCurrentMonth) ? toTimestamp(endOfCurrentMonth) : this.props.highlightedIndex)
                : toTimestamp(highlightedIndex);

            onHighlightedIndexChange(toTimestamp(currentMonth), hi);
        },

        collapse() {
            if (isHandheld()) {
                return;
            }

            const currentMonth = this.props.value !== null
                ? toTimestamp(moment(this.props.value).startOf('month'))
                : +new Date();

            const highlightedIndex = this.props.value !== null
                ? toTimestamp(moment(this.props.value))
                : +new Date();

            this.props.onCollapse(currentMonth, highlightedIndex);
        },

        expand() {
            if (isHandheld()) {
                return;
            }

            this.props.onExpand();
        },

        focus() {
            ReactDOM.findDOMNode(this.component).focus();
        },

        isDisabledDate(date) {
            const value = toTimestamp(date);
            return this.props.isDisabledDate(value);
        },

        moveToNextMonth() {
            const { currentMonth, highlightedIndex } = this.props;
            this.changeHighlightedIndex(
                moment(currentMonth).add(1, 'months'),
                moment(highlightedIndex).add(1, 'months'),
                true
            );
        },

        moveToPreviousMonth() {
            const { currentMonth, highlightedIndex } = this.props;
            this.changeHighlightedIndex(
                moment(currentMonth).subtract(1, 'months'),
                moment(highlightedIndex).subtract(1, 'months'),
                true
            );
        },

        renderCalendar() {
            const { currentMonth, expanded, hasGrayedOutPast, highlightedIndex, tetherTo, value } = this.props;
            if (!expanded || isHandheld()) {
                return null;
            }

            return (
                <FoundationTether
                    target={ () => ReactDOM.findDOMNode(this) }
                    tetherTo={ tetherTo }>
                    <Calendar
                        currentMonth={ moment(currentMonth) }
                        hasGrayedOutPast={ hasGrayedOutPast }
                        isDisabledDate={ this.isDisabledDate }
                        onNextMonthClick={ this.onNextMonthClick }
                        onPreviousMonthClick={ this.onPreviousMonthClick }
                        onChange={ this.onChange }
                        highlightedIndex={ moment(highlightedIndex) }
                        value={ moment(value) } />
                </FoundationTether>
            );
        },

        render() {
            const { disabled, expanded, ui } = this.props;
            const touchable = isHandheld();

            const value = this.props.value !== null
                ? moment(this.props.value).format('YYYY-MM-DD')
                : this.props.emptyText;

            return (
                <div
                    className={ block({ disabled, expanded, touchable }, uiPropGetValue(ui)) }
                    onKeyDown={ this.onKeyDown }
                    onClick={ this.onToggle }
                    ref={ (c) => this.component = c }
                    tabIndex={ disabled ? null : '0' }>
                    <Control
                        defaultValue={ this.props.value !== null ? value : null }
                        disabled={ disabled }
                        onTouchableBlur={ this.onCollapse }
                        onTouchableChange={ this.onTouchableChange }
                        onTouchableFocus={ this.onExpand }
                        value={ value } />
                    { this.renderCalendar() }
                </div>
            );
        },

    }));

};
